import pytest
import time
import math
import select
import fiber
import docker
import requests
from fiber import Process
from fiber.docker_backend import Backend

import fiber.config as fiber_config


def func_add(a, b):
    return a + b


def func_error():
    raise ValueError


@pytest.fixture(scope="module")
def client():
    return docker.from_env()


class TimeoutBackend(Backend):
    def __init__(self, n):
        # the first n times calling create_job will result a timeout
        super(TimeoutBackend, self).__init__()
        self.count = 0
        self.n = n

    def create_job(self, job_spec):
        self.count += 1
        if self.count <= self.n:
            raise requests.exceptions.ReadTimeout
        else:
            return super(TimeoutBackend, self).create_job(job_spec)


def square_worker(a):
    return a**2


class TestBasic():
    @pytest.fixture(autouse=True)
    def check_leak(self, request, client):
        assert fiber.active_children() == []
        if fiber.config.default_backend != "docker":
            yield
            return
        pre_workers = client.containers.list()
        yield
        post_workers = client.containers.list()
        assert pre_workers == post_workers

    def test_process(self):
        p = Process(target=func_add, args=(42, 89), name="test_process")
        p.start()
        p.join()
        assert 0 == p.exitcode

    def test_process_error(self):
        p = Process(target=func_error, name="test_process_error")
        p.start()
        p.join()
        assert 0 != p.exitcode

    def test_none_process(self):
        p = Process(target=None, name="test_none_process")
        p.start()
        p.join()
        assert 0 == p.exitcode

    def test_multi_processes(self):
        n = 3
        procs = []
        for i in range(n):
            procs.append(Process(target=func_add, args=(42, 89), name="test_multi_processes"))
        for p in procs:
            p.start()
        for p in procs:
            p.join()
        for p in procs:
            assert 0 == p.exitcode

    def test_process_run(self):
        p = Process(target=func_error, name="test_process_run")
        with pytest.raises(ValueError):
            p.run()

    def test_join_timeout(self):
        p = Process(target=func_add, args=(42, 89), name="test_join_timeout")
        p.start()
        now = time.time()
        p.join(1)
        diff = time.time() - now
        # because docker backend status updates are every 1 second. In the
        # worst case, this should be 1 second
        # make sure there is no leaked processes
        p.join()
        assert math.fabs(diff - 1) < 1

    def test_is_alive(self):
        p = Process(target=func_add, args=(42, 89), name="test_is_alive")
        assert p.is_alive() is False
        p.start()
        assert p.is_alive() is True
        p.join()
        assert p.is_alive() is False

    def test_name(self):
        p = Process(target=func_add, args=(42, 89), name="my_process")
        assert p.name == "my_process"

    def test_daemon(self):
        p = Process(target=func_add, args=(42, 89), daemon=True, name="test_daemon")
        p.start()
        assert p.daemon is True
        p.join()

    def test_pid(self):
        p = Process(target=func_add, args=(42, 89), daemon=True, name="test_pid")
        assert p.pid is None
        p.start()
        assert p.pid is not None
        assert p.pid < 32768
        assert p.pid > 0
        p.join()

    def test_authkey(self):
        p = Process(target=func_add, args=(42, 89), daemon=True, name="test_authkey")
        assert p.authkey is not None

    def test_sentinel(self):
        p = Process(target=time.sleep, args=(0.2,), name="test_sentinel")
        p.start()
        fds = select.select([p.sentinel], [], [], 0.1)
        assert len(fds[0]) == 0
        fds = select.select([p.sentinel], [], [])
        assert len(fds[0]) == 1
        assert fds[0][0] == p.sentinel
        p.join()

    def test_terminate(self):
        p = Process(target=time.sleep, args=(1000,), name="test_terminate")
        p.start()
        assert p.is_alive() is True
        p.terminate()
        time.sleep(0.5)
        assert p.is_alive() is False
        # when terminated, exit code can be 0 for k8s backend
        #assert p.exitcode != 0
        p.join()

    def test_exitcode(self):
        p = Process(target=func_add, args=(42, 89), name="test_terminate")
        p.start()
        while True:
            if p.exitcode is not None:
                break
            time.sleep(1)
        assert p.exitcode == 0

    def test_ipc_passive(self):
        if fiber.config.default_backend == "kubernetes":
            # ipc passive is not very useful for current setup, consider disable this completely
            pytest.skip("skipped because current backend is kubernetes")
        # worker process waits for connection from master process
        fiber_config.ipc_active = False
        try:
            p = Process(target=func_add, args=(42, 89), name="test_ipc_passive")
            p.start()
            p.join()
            assert 0 == p.exitcode
        finally:
            fiber_config.ipc_active = True

    def test_start_timeout(self):
        fiber.backend.get_backend(name="docker")
        old_backend = fiber.backend._backends["docker"]
        fiber.backend._backends["docker"] = TimeoutBackend(n=4)

        p = fiber.Pool(4)
        res = p.map(square_worker, [1, 2, 3, 4])
        p.terminate()
        fiber.backend._backends["docker"] = old_backend

        assert res == [i**2 for i in range(1, 5)]

        # wait for 2 seconds to let docker finish starting
        #time.sleep(2)

# name, daemon, pid, authkey, sentinel, terminate



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import importlib
import multiprocessing as mp
import logging
import fiber.config as config
logger = logging.getLogger('Trax')


_backends = {}
available_backend = ['kubernetes', 'docker', 'local']


def is_inside_kubenetes_job():
    if os.environ.get("KUBERNETES_SERVICE_HOST", None):
        return True
    return False


def is_inside_docker_job():
    if os.environ.get("FIBER_BACKEND", "") == "docker":
        return True
    return False


BACKEND_TESTS = {
    "kubernetes": is_inside_kubenetes_job,
    "docker": is_inside_docker_job,
}


def auto_select_backend():
    for backend_name, test in BACKEND_TESTS.items():
        if test():
            name = backend_name
            break
    else:
        name = config.default_backend

    return name


def get_backend(name=None, **kwargs):
    """
    Returns a working Fiber backend. If `name` is specified, returns a
    backend specified by `name`.
    """
    global _backends
    if name is None:
        if config.backend is not None:
            name = config.backend
        else:
            name = auto_select_backend()
    else:
        if name not in available_backend:
            raise mp.ProcessError("Invalid backend: {}".format(name))

    _backend = _backends.get(name, None)
    if _backend is None:
        _backend = importlib.import_module("fiber.{}_backend".format(
            name)).Backend(**kwargs)
        _backends[name] = _backend
    logger.debug(f'backend name is {name}')
    return _backend



import re
import io
import os
import logging
import glob
import fiber
import fiber.config
from fiber.docker_backend import Backend
import time
import pytest
import multiprocessing
from unittest import mock

import fiber.config as fiber_config

def pytest_sessionstart(session):
    assert 1 == 0
    print("session start")


def pytest_sessionfinish(session, exitstatus):
    assert 1 == 0
    print("session finished")


def import_missing_module_worker():
    import this_module_does_not_exist


def dummy_q_worker(q):
    q.put("DUMMY_WORKER_DONE")
    time.sleep(0.1)


class TestMisc():
    backend = Backend()

    # Throw an error so that the job won't be started. `python:3.6-alpine`
    # doesn't have proper depencency installed like `pytest`
    @mock.patch.object(fiber.docker_backend.Backend, 'create_job',
                       side_effect=NotImplementedError)
    def test_init_image(self, mock_create_job):
        if fiber.config.default_backend != "docker":
            pytest.skip("skipped because current backend is not docker")

        try:
            fiber.init(image='python:3.6-alpine', backend='docker')
            p = fiber.Process(name="test_init_image")
            p.start()
        except NotImplementedError:
            pass
        finally:
            fiber.reset()
        args = mock_create_job.call_args
        # https://docs.python.org/3/library/unittest.mock.html#unittest.mock.Mock.call_args # noqa E501
        job_spec_arg = args[0][0]
        assert job_spec_arg.image == "python:3.6-alpine"

    @mock.patch.object(fiber.docker_backend.Backend, 'create_job',
                       side_effect=NotImplementedError)
    def test_init_image2(self, mock_create_job):
        if fiber.config.default_backend != "docker":
            pytest.skip("skipped because current backend is not docker")

        # don't set backend when init

        try:
            fiber.init(image='python:3.6-alpine')
            p = fiber.Process(name="test_init_image2")
            p.start()
        except NotImplementedError:
            pass
        finally:
            fiber.reset()
        args = mock_create_job.call_args
        # https://docs.python.org/3/library/unittest.mock.html#unittest.mock.Mock.call_args # noqa E501
        job_spec_arg = args[0][0]
        assert job_spec_arg.image == "python:3.6-alpine"

    def test_interactive_shell_cloudpickle(self):
        def add(a, b):
            return a + b

        # This will fail: "AttributeError: Can't pickle local object"
        # default pickler can pickle local object
        with pytest.raises(AttributeError):
            p = fiber.Process(target=add, args=(1, 2),
                              name="test_interactive_shell_cloudpickle")
            p.start()

        # This will work. When using interactive console, pickler
        # is set to cloudpickle

        with mock.patch("fiber.util.is_in_interactive_console") as func:
            func.return_value = True

            p = fiber.Process(target=add, args=(1, 2),
                              name="test_interactive_shell_cloudpickle2")
            p.start()
            p.join()

    def test_image_not_found(self):
        if fiber.config.default_backend != "docker":
            pytest.skip("skipped because current backend is not docker")

        try:
            with pytest.raises(multiprocessing.ProcessError):
                fiber.init(
                    image='this-image-does-not-exist-and-is-only-used-for-testing'
                )
                p = fiber.Process(name="test_image_not_found")
                p.start()
        finally:
            fiber.reset()

    """
    def test_push_pull_queue(self):
        # Use pull queue by default
        fiber.reset()
        q = fiber.SimpleQueue()
        assert isinstance(q, fiber.queues.SimpleQueuePull)

        try:
            fiber.init(use_push_queue=True)
            q = fiber.SimpleQueue()
            assert isinstance(q, fiber.queues.SimpleQueuePush)
        finally:
            fiber.reset()

        try:
            fiber.init(use_push_queue=False)
            q = fiber.SimpleQueue()
            assert isinstance(q, fiber.queues.SimpleQueuePull)
        finally:
            fiber.reset()
    """

    # Instead of raising an exception, Fiber now just print an error log
    def test_no_python3_inside_image(self):
        if fiber.config.default_backend != "docker":
            pytest.skip("skipped because current backend is not docker")

        '''
        fp = io.StringIO()
        handler = logging.StreamHandler(fp)
        logger = logging.getLogger("fiber")
        logger.setLevel(level=logging.DEBUG)
        logger.addHandler(handler)
        '''

        with pytest.raises(multiprocessing.ProcessError):
            try:
                fiber.init(image='ubuntu:18.04')
                p = fiber.Process(name="test_no_python3_inside_image")
                p.start()
            finally:
                fiber.reset()
                #logger.removeHandler(handler)

        '''
        try:
            fiber.init(image='ubuntu:18.04')
            p = fiber.Process(name="test_no_python3_inside_image")
            p.start()

            logs = fp.getvalue()
            times = len(list(re.finditer('Failed to start Fiber process',
                                         logs)))
            assert times == 1
        finally:
            fiber.reset()
            logger.removeHandler(handler)
        '''

    """
    def test_missing_module_inside_image(self):
        with pytest.raises(multiprocessing.ProcessError):
            p = fiber.Process(target=import_missing_module_worker, name="test_missing_module_inside_image")
            p.start()
    """

    def test_log_file_path(self):
        if fiber.config.default_backend != "docker":
            pytest.skip("skipped because current backend is not docker")

        # make sure workers and master doesn't write to the same log file
        try:
            log_file = fiber_config.log_file
            log_level = fiber_config.log_level
            # clean up
            if os.path.isdir("tests/logs"):
                files = glob.glob("tests/logs/*")
                for f in files:
                    os.remove(f)
            files = glob.glob("tests/logs/*")
            assert files == []

            fiber.init(log_file="tests/logs/fiber.log", log_level="DEBUG")
            q = fiber.SimpleQueue()
            p = fiber.Process(target=dummy_q_worker, args=(q,))
            p.start()
            msg = q.get()

            assert msg == "DUMMY_WORKER_DONE"

            files = glob.glob("tests/logs/*")
            files.sort()
            assert len(files) == 2
            assert files[0] == "tests/logs/fiber.log.MainProcess"
            assert "tests/logs/fiber.log.Process-" in files[1]

            regex = r'\w+:Process-\d+\(\d+\):'
            with open("tests/logs/fiber.log.MainProcess") as f:
                content = f.read()
                assert re.search(regex, content) is None, (
                    "Fiber subprocess log found in master log file")

            p.join()

        finally:
            fiber.init(log_file=log_file, log_level=log_level)



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fiber introduces a new concept called *job-backed processes*. It is similar to
the *process* in Python's multiprocessing library, but more flexible: while a
process in multiprocessing only runs on a local machine, a Fiber process can
run remotely on a different machine or locally on the same machine.

When starting a new Fiber process, Fiber creates a new job with the proper Fiber
backend on the current computer cluster. Fiber uses containers to encapsulate
the running environment of current processes, including all the required files,
input data, and other dependent program packages, etc., to ensure everything is
self-contained. All the child processes are started with the same container
image as the parent process to guarantee a consistent running environment.

Because each process is a cluster job, its life cycle is the same as any job on
the cluster.
"""

import sys
import fiber.util
import itertools
import multiprocessing as mp
import logging

from multiprocessing.process import BaseProcess


logger = logging.getLogger('Trax')


_children = set()
_current_process = mp.current_process()


def _cleanup():
    # check for processes which have finished
    for p in list(_children):
        if p._popen.poll() is not None:
            _children.discard(p)


def active_children():
    """
    Get a list of children processes of the current process.

    :returns: A list of children processes.

    Example:
    ```python
    p = fiber.Process(target=time.sleep, args=(10,))
    p.start()
    print(fiber.active_children())
    ```
    """
    _cleanup()
    return list(_children)


def current_process():
    """Return a Process object representing the current process.

    Example:
    ```python
    print(fiber.current_process())
    ```
    """
    return _current_process


class Process(BaseProcess):
    """
    Create and manage Fiber processes.

    The API is compatible with Python's multiprocessing.Process, check
    [here](https://docs.python.org/3.6/library/multiprocessing.html#process-and-exceptions)
    for multiprocessing's documents.

    Example usage:

    ```python
    p = Process(target=f, args=('Fiber',))
    p.start()
    ```

    ---

    ###pid

    ```
    Process.pid
    ```

    Return the current process PID. If the process hasn't been fully started,
    the value will be `None`. This PID is assigned by Fiber and is different
    from the operating system process PID. The value of `pid` is derived from
    the job ID of the underlying job that runs this Fiber process.

    ###name
    ```
    Process.name
    ```

    Return the name of this Fiber process. This value need to be set before the
    start of the Process.

    Example:
    ```python
    p = fiber.Process(target=print, args=("demo"))
    p.name = "DemoProcess"
    ```

    ###daemon

    ```
    Process.daemon
    ```

    In multiprocessing, if a process has daemon set, it will be terminated when
    it's parent process exits. In Fiber, current this value has no effect, all
    processes will be cleaned up when their parent exits.

    ###authkey

    ```
    Process.authkey
    ```
    Authkey is used to authenticate between parent and child processes. It is
    a byte string.

    ###exitcode

    ```
    Process.exitcode
    ```
    The exit code of current process. If the process has not exited, the
    exitcode is `None`. If the current process has exited, the value of this
    is an integer.

    ###sentinel

    ```
    Process.sentinel
    ```
    Returns a file descriptor that becomes "ready" when the process exits. You
    can call `select` and other eligible functions that works on fds on this
    file descriptor.
    """
    _start_method = None
    _pid = None

    @staticmethod
    def _Popen(process_obj):
        from .popen_fiber_spawn import Popen
        return Popen(process_obj)

    def __init__(self, group=None, target=None, name=None, args=(), kwargs={},
                 *, daemon=None):
        super(Process, self).__init__(group=group, target=target, name=name,
                                      args=args, kwargs=kwargs, daemon=daemon)
        self._parent_pid = current_process().pid
        # set when Process.start() failed
        self._start_failed = False

    def __repr__(self):
        return "{}({}, {})>".format(type(self).__name__, self._name, self.ident)

    def run(self):
        """Run the target function of current process (in current process)

        :returns: return value of the target function
        """
        return super().run()

    def start(self):
        """Start this process.

        Under the hood, Fiber calls the API on the computer cluster to start a
        new job and run the target function in the new job.
        """
        assert self._popen is None, 'cannot start a process twice'
        assert self._parent_pid == current_process().pid, \
            'can only start a process object created by current process'
        assert not _current_process._config.get('daemon'), \
            'daemonic processes are not allowed to have children'
        _cleanup()

        _popen = self._Popen(self)
        # put waiting logic in another function so that self._popen can be set
        # earlier. This make sure that clean up code can find self._popen.
        logger.debug("create new process %s", self)
        _popen._launch(self)

        # This is not needed because Process._popen is set inside _launch()
        # function
        # self._popen = _popen
        logger.debug("process %s, set sentinel %s",
                     self, self._popen.sentinel)
        self._sentinel = self._popen.sentinel
        # Avoid a refcycle if the target function holds an indirect
        # reference to the process object (see bpo-30775)
        del self._target, self._args, self._kwargs
        _children.add(self)

    def terminate(self):
        """Terminate current process.

        When running locally, Fiber sends an SIGTERM signal to the child
        process. When running on a computer cluster, Fiber calls the
        corresponding API on that platform to terminate the job that runs Fiber
        process.
        """
        if self._popen is None:
            logger.debug("self._popen is None when terminate() is called")
            return
        self._popen.terminate()

    def join(self, timeout=None):
        """Wait for this process to terminate.

        :param timeout: The maximum duration of time in seconds that this call
            should wait before return. if `timeout` is `None` (default value),
            this method will wait forever until the process terminates. If
            `timeout` is `0`, it will check if the process has exited and
            return immediately.

        :returns: The exit code of this process
        """
        return super().join(timeout=timeout)

    def is_alive(self):
        """Check if current process is still alive

        :returns: `True` if current process is still alive. Returns `False` if
            it's not alive.
        """
        return super().is_alive()

    @property
    def ident(self):
        if self._pid is None:
            self._pid = self._popen and self._popen.pid

        return self._pid

    @ident.setter
    def ident(self, pid):
        self._pid = pid

    pid = ident

    def _bootstrap(self):
        from multiprocessing import util, context
        global _current_process, _process_counter, _children
        err = None

        try:
            if self._start_method is not None:
                context._force_start_method(self._start_method)
            _process_counter = itertools.count(1)
            _children = set()
            util._close_stdin()
            old_process = _current_process
            _current_process = self

            try:
                fiber.util._finalizer_registry.clear()
                fiber.util._run_after_forkers()
            except Exception as e:
                err = e
            finally:
                # delay finalization of the old process object until after
                # _run_after_forkers() is executed
                del old_process
            util.info('child process calling self.run()')
            try:
                self.run()
                exitcode = 0
            except Exception as e:
                import traceback
                err = traceback.format_exc()
                exitcode = 1
            finally:
                util._exit_function()
        except SystemExit as e:
            err = None
            if not e.args:
                exitcode = 0
            elif isinstance(e.args[0], int):
                exitcode = e.args[0]
            else:
                sys.stderr.write(str(e.args[0]) + '\n')
                exitcode = 1
        except Exception as e: # noqa E722
            err = e
            exitcode = 1
            import traceback
            msg = traceback.format_exc()
            sys.stderr.write('Process {} exception: {}\n'.format(
                self.name, msg))
        finally:
            try:
                sys.stdout.flush()
            except (AttributeError, ValueError):
                pass
            try:
                sys.stderr.flush()
            except (AttributeError, ValueError):
                pass

        return exitcode, err



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Local Fiber backend helps Fiber to run locally on a single machine.
This backend is mainly used for fast prototyping.
"""
import subprocess

import fiber.core as core
from fiber.core import ProcessStatus


class Backend(core.Backend):
    """
    We use subprocess to simulate job launch locally. Python's subprocess.Popen
    has already provided us all the necessary APIs to control subprocess, we
    just need to call them here.
    """
    name = "local"

    def __init__(self):
        pass

    def create_job(self, job_spec):
        proc = subprocess.Popen(job_spec.command)
        job = core.Job(proc, proc.pid)
        job.host = 'localhost'

        return job

    def get_job_status(self, job):
        proc = job.data

        if proc.poll() is not None:
            # subprocess has terminated
            return ProcessStatus.STOPPED

        return ProcessStatus.STARTED

    def wait_for_job(self, job, timeout):
        proc = job.data

        if timeout == 0:
            return proc.poll()

        try:
            proc.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            return None

        return proc.returncode

    def terminate_job(self, job):
        proc = job.data

        proc.terminate()

    def get_listen_addr(self):
        return "127.0.0.1", 0, "lo"



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import logging
import multiprocessing as mp
import threading

from fiber.backend import get_backend


MIN_PORT = 40000
MAX_PORT = 65535
logger = logging.getLogger("fiber")

socket_lib = "nanomsg"
default_socket_ctx = None


if socket_lib == "nanomsg":
    import nnpy
    import nnpy.errors
elif socket_lib == "nng":
    import pynng
    from pynng.exceptions import check_err
elif socket_lib == "zmq":
    import zmq
    import zmq.devices
else:
    raise ValueError("bad socket_lib value: {}".format(socket_lib))


def get_ctx():
    return default_socket_ctx


def bind_to_random_port(sock, addr_base, max_tries=100, nng=True):
    num_tries = 0
    while num_tries < max_tries:
        try:
            port = random.randint(MIN_PORT, MAX_PORT)
            addr = "{}:{}".format(addr_base, port)
            if nng:
                sock.listen(addr)
            else:
                sock.bind(addr)
            return port
        except:
            num_tries += 1
            continue

    return


class SockContext:
    default_addr = None

    def new(self, mode):
        raise NotImplementedError

    @staticmethod
    def bind_random(sock, addr):
        raise NotImplementedError

    @staticmethod
    def connect(sock, addr):
        raise NotImplementedError

    @staticmethod
    def close(sock):
        sock.close()


class ZMQContext(SockContext):
    default_addr = "tcp://0.0.0.0"

    def __init__(self):

        self._mode_to_type = {
            "r": zmq.DEALER,
            "w": zmq.DEALER,
            "rw": zmq.DEALER,
            "req": zmq.REQ,
            "rep": zmq.REP,
        }
        self.context = zmq.Context.instance()

    def new(self, mode):
        sock_type = self._mode_to_type[mode]
        if sock_type is None:
            return None
        return self.context.socket(sock_type)

    @staticmethod
    def bind_random(sock, addr):
        assert type(addr) == str
        return sock.bind_to_random_port(
            addr, min_port=MIN_PORT, max_port=MAX_PORT, max_tries=100
        )

    @staticmethod
    def connect(sock, addr):
        return sock.connect(addr)

    def device(self, s1_mode, s2_mode):

        backend = get_backend()
        ip_ext, _, _ = backend.get_listen_addr()
        ip_bind = "0.0.0.0"
        addr_bind = "tcp://{}".format(ip_bind)

        s1_type = self._mode_to_type[s1_mode]
        s2_type = self._mode_to_type[s2_mode]

        device = zmq.devices.ThreadDevice(in_type=s1_type, out_type=s2_type)
        _writer_port = device.bind_in_to_random_port(
            addr_bind, min_port=MIN_PORT, max_port=MAX_PORT, max_tries=100
        )
        _reader_port = device.bind_out_to_random_port(
            addr_bind, min_port=MIN_PORT, max_port=MAX_PORT, max_tries=100
        )
        _reader_ext_addr = "tcp://{}:{}".format(ip_ext, _reader_port)
        _writer_ext_addr = "tcp://{}:{}".format(ip_ext, _writer_port)
        return device, _reader_ext_addr, _writer_ext_addr


class NNGDevice:
    STATE_INIT = 0
    STATE_AFTER_INIT = 1
    STATE_FINISHED = 2

    def __init__(self, ctx, s1_mode, s2_mode, default_addr="tcp://0.0.0.0"):

        self._mode_to_opener = {
            "r": pynng.lib.nng_pull0_open_raw,
            "w": pynng.lib.nng_push0_open_raw,
            "rw": None,
        }

        self.s1_mode = s1_mode
        self.s2_mode = s2_mode
        self.default_addr = default_addr
        self._start_process()

    def _start_process(self):
        parent_conn, child_conn = mp.Pipe()
        self._proc = threading.Thread(
            target=self._run, args=(child_conn,), daemon=True
        )
        '''
        self._proc = mp.Process(
            target=self._run, args=(child_conn,), daemon=True
        )
        '''
        self._proc.start()

        #child_conn.close()
        self.conn = parent_conn

    def _mode_to_opener(self, mode):
        opener = self._mode_to_opener[mode]
        if opener is None:
            raise ValueError(
                "Mode {} not supported by {}", mode, self.__class__.__name__
            )
        return opener

    def _create_socks(self):

        opener1 = self._mode_to_opener[self.s1_mode]
        opener2 = self._mode_to_opener[self.s2_mode]

        s1 = pynng.Socket(opener=opener1)
        s2 = pynng.Socket(opener=opener2)

        return s1, s2

    def _bind_socks(self, s1, s2):
        port1 = bind_to_random_port(s1, self.default_addr)
        port2 = bind_to_random_port(s2, self.default_addr)

        return port1, port2

    def _run_device(self, s1, s2):

        ret = pynng.lib.nng_device(s1.socket, s2.socket)
        check_err(ret)

    def _run(self, conn):
        s1, s2 = self._create_socks()

        state = NNGDevice.STATE_INIT

        while True:
            cmd = conn.recv()
            if state == NNGDevice.STATE_INIT:
                # wait for bind
                if cmd != "#bind":
                    conn.send("bad command, expecting #bind")
                else:
                    logger.debug("%s got #bind", self.__class__.__name__)
                    # bind sockets
                    port1, port2 = self._bind_socks(s1, s2)

                    conn.send((port1, port2))

                    state = NNGDevice.STATE_AFTER_INIT
            elif state == NNGDevice.STATE_AFTER_INIT:
                # wait for start()
                if cmd != "#start":
                    conn.send("bad command, expecting #start")
                else:
                    logger.debug("%s got #start", self.__class__.__name__)
                    # ACK
                    conn.send(None)
                    # Keeps running forever
                    self._run_device(s1, s2)
                    state = NNGDevice.STATE_FINISHED
                    logger.debug("%s finished", self.__class__.__name__)
                    conn.send("bad command, expecting #start")
                    break
            else:
                break

    def bind(self):
        self.conn.send("#bind")
        in_addr, out_addr = self.conn.recv()
        if in_addr is None:
            raise RuntimeError("NNGDevice failed to bind")
        self.in_addr = in_addr
        self.out_addr = out_addr
        return in_addr, out_addr

    def start(self):
        self.conn.send("#start")
        code = self.conn.recv()
        if code is not None:
            raise RuntimeError("Failed to start NNGDevice, {}".format(code))


class NNGContext(SockContext):
    default_addr = "tcp://0.0.0.0"

    def __init__(self):

        self._mode_to_creator = {
            "r": pynng.Pull0,
            "w": pynng.Push0,
            "rw": pynng.Pair0,
            "req": pynng.Req0,
            "rep": pynng.Rep0,
        }

    def new(self, mode):
        func = self._mode_to_creator[mode]
        if func is None:
            return None
        return func()

    @staticmethod
    def bind_random(sock, addr):
        return bind_to_random_port(sock, addr)

    @staticmethod
    def connect(sock, addr):
        return sock.dial(addr)

    def device(self, s1_mode, s2_mode):
        self.s1_mode = s1_mode
        self.s2_mode = s2_mode

        device = NNGDevice(
            self, s1_mode, s2_mode, default_addr=NNGContext.default_addr
        )
        in_port, out_port = device.bind()

        backend = get_backend()
        ip_ext, _, _ = backend.get_listen_addr()

        in_addr = "tcp://{}:{}".format(ip_ext, in_port)
        out_addr = "tcp://{}:{}".format(ip_ext, out_port)

        return device, in_addr, out_addr


class NanomsgDevice(NNGDevice):
    def __init__(self, ctx, s1_mode, s2_mode, default_addr="tcp://0.0.0.0"):
        self.s1_mode = s1_mode
        self.s2_mode = s2_mode
        self.default_addr = default_addr
        self.ctx = ctx

        self._start_process()

    def _create_socks(self):

        s1 = nnpy.Socket(nnpy.AF_SP_RAW, self.ctx._mode_to_type[self.s1_mode])
        s2 = nnpy.Socket(nnpy.AF_SP_RAW, self.ctx._mode_to_type[self.s2_mode])
        return s1, s2

    def _bind_socks(self, s1, s2):
        port1 = bind_to_random_port(s1, self.default_addr, nng=False)
        port2 = bind_to_random_port(s2, self.default_addr, nng=False)
        return port1, port2

    def _run_device(self, s1, s2):

        rc = nnpy.nanomsg.nn_device(s1.sock, s2.sock)
        return nnpy.errors.convert(rc, rc)


class NanomsgContext(SockContext):
    default_addr = "tcp://0.0.0.0"

    def __init__(self):

        self._mode_to_type = {
            "r": nnpy.PULL,
            "w": nnpy.PUSH,
            "rw": nnpy.PAIR,
            "rep": nnpy.REP,
            "req": nnpy.REQ,
        }

    def new(self, mode):

        sock_type = self._mode_to_type[mode]
        if sock_type is None:
            return None
        return nnpy.Socket(nnpy.AF_SP, sock_type)

    @staticmethod
    def bind_random(sock, addr):
        return bind_to_random_port(sock, addr, nng=False)

    @staticmethod
    def connect(sock, addr):
        return sock.connect(addr)

    def device(self, s1_mode, s2_mode):
        self.s1_mode = s1_mode
        self.s2_mode = s2_mode

        device = NanomsgDevice(
            self, s1_mode, s2_mode, default_addr=NanomsgContext.default_addr
        )
        in_port, out_port = device.bind()

        backend = get_backend()
        ip_ext, _, _ = backend.get_listen_addr()

        in_addr = "tcp://{}:{}".format(ip_ext, in_port)
        out_addr = "tcp://{}:{}".format(ip_ext, out_port)

        return device, in_addr, out_addr


if socket_lib == "nanomsg":
    default_socket_ctx = NanomsgContext()
elif socket_lib == "nng":
    default_socket_ctx = NNGContext()
elif socket_lib == "zmq":
    default_socket_ctx = ZMQContext()
else:
    raise ValueError("bad socket_lib value: {}".format(socket_lib))


class Socket:
    def __repr__(self):
        return "{}<{},{}>".format(
            self.__class__.__name__,
            self._ctx.__class__.__name__,
            self._mode)

    def __init__(self, ctx=get_ctx(), mode="rw"):
        self._mode = mode
        self._ctx = ctx
        self._sock = ctx.new(mode)
        if self._sock is None:
            raise ValueError("Socket mode \"{}\" not supported by {}".format(
                mode, ctx.__class__.__name__))

    def send(self, data):
        self._sock.send(data)

    def recv(self):
        return self._sock.recv()

    def bind(self):
        addr = self._ctx.default_addr
        bind_random = self._ctx.bind_random
        port = bind_random(self._sock, addr)

        return port

    def connect(self, addr):
        _connect = self._ctx.connect
        _connect(self._sock, addr)

    def close(self):
        _close = self._ctx.close
        _close(self._sock)


class ProcessDevice:
    def __init__(self, s1_mode, s2_mode, ctx=get_ctx()):
        device, in_addr, out_addr = ctx.device(s1_mode, s2_mode)
        self.device = device
        self.in_addr = in_addr
        self.out_addr = out_addr

    def start(self):
        self.device.start()
        logger.debug("started device in:%s out:%s", self.in_addr, self.out_addr)



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module deals with Fiber configurations.

There are 3 way of setting Fiber configurations: config file, environment
variable and Python code. The priorities are: Python code > environment
variable > config file.

#### Config file

Fiber config file is a plain text file following Python's
[configparser](https://docs.python.org/3.6/library/configparser.html) file
format. It needs to be named `.fiberconfig` and put into the directory where
you launch your code.

An example `.fiberconfig` file:
```
[default]
log_level=debug
log_file=stdout
backend=local
```

#### Environment variable

Alternatively, you can also use environment variables to pass configurations to
Fiber. The environment variable names are in format `FIBER_` + config name in
upper case.

For example, an equivalent way of specifying the above config using environment
variables is:

```
FIBER_LOG_LEVEL=debug FIBER_LOG_FILE=stdout FIBER_BACKEND=local python code.py ...
```

#### Python code

You can also set Fiber config in your Python code:

```python
import fiber.config as fiber_config
...
def main():
    fiber_config.log_level = "debug"
    fiber_config.log_file = "stdout"
    fiber_config.backend = "local"
```

Note that almost all of the configurations needs to be set before you launch
any Fiber processes.
"""

import os
import logging
import configparser


_current_config = None
logger = logging.getLogger('Trax')


LOG_LEVELS = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
}

DEFAULT_IMAGE = "fiber-test:latest"


def str2bool(text):
    """Simple function to convert a range of values to True/False."""
    return text.lower() in ["true", "yes", "1"]


class Config(object):
    """Fiber configuration object. Available configurations:

    | key                   | Type | Default | Notes |
    | --------------------- |:------|:-----|:------|
    | debug                 | bool  | False | Set this to `True` to turn on debugging |
    | image                 | str   | None  | Docker image to use when starting  new processes |
    | default_image         | str   | None  | Default docker image to use when `image` config value  is not set |
    | backend               | str   | None  | Fiber backend to use when starting new processes. Check [here](platforms.md) for available backends |
    | default_backend       | str   | `local` | Default Fiber backend to use when `backend` config is not set |
    | log_level             | str/int | `logging.INFO` | Fiber log level. This config accepts either a int value (log levels from `logging` module like `logging.INFO`) or strings: `debug`, `info`, `warning`, `error`, `critical` |
    | log_file              | str   | `/tmp/fiber.log` | Default fiber log file path. Fiber will append the process name to this value and create one log file for each process. A special value `stdout` means to print the logs to standard output |
    | ipc_admin_master_port | int   | `0`   | The port that master process uses to communicate with child processes. Default value is `0` which means the master process will choose a random port |
    | kubernetes_namespace | str   | `default` | The namespace that Fiber `kubernetes` backend will use to create pods and do other work on Kubernetes |


    """
    def __init__(self, conf_file=None):
        # Not documented, people should not use this
        self.merge_output = False
        self.debug = False
        self.image = None
        self.default_image = DEFAULT_IMAGE
        self.backend = None
        self.default_backend = "local"
        # Not documented, this should be removed because it's not used for now
        self.use_bash = False
        self.log_level = logging.INFO
        self.log_file = "stdout"
        # If ipc_active is True, Fiber worker processes will connect
        # to the master process. Otherwise, the master process will connect
        # to worker processes.
        # Not documented, should only be used internally
        self.ipc_active = True
        # if ipc_active is True, this can be 0, otherwise, it can only be a
        # valid TCP port number. Default 0.
        self.ipc_admin_master_port = 0
        # Not documented, this is only used when `ipc_active` is False
        self.ipc_admin_worker_port = 8000
        # Not documented, need to fine tune this
        self.cpu_per_job = 1
        # Not documented, need to fine tune this
        self.mem_per_job = None
        self.use_push_queue = True
        self.kubernetes_namespace = "default"

        if conf_file is None:
            conf_file = ".fiberconfig"

        # Load config from config file
        if os.path.exists(conf_file):
            logger.debug("loading config from %s", conf_file)
            config = configparser.ConfigParser()
            config.read(conf_file)
            for k in config["default"]:
                if k in self.__dict__:
                    self.__dict__[k] = config["default"][k]
                else:
                    raise ValueError(
                        'unknown config key "{}" in {}. Valid keys: '
                        '{}'.format(k, conf_file,
                                    [key for key in self.__dict__]))

        else:
            logger.debug("no fiber config file (%s) found", conf_file)

        # load environment variable overwrites
        for k in self.__dict__:
            name = "FIBER_" + k.upper()
            val = os.environ.get(name, None)
            if val:
                self.__dict__[k] = val

        # rewrite values
        if isinstance(self.log_level, str):
            level = self.log_level.lower()
            if level not in LOG_LEVELS:
                logger.debug("bad logging level: %s", self.log_level)
                level = logging.NOTSET
            else:
                level = LOG_LEVELS[level]
            self.log_level = level

        if isinstance(self.ipc_active, str):
            self.ipc_active = str2bool(self.ipc_active)

        if isinstance(self.cpu_per_job, str):
            self.cpu_per_job = int(self.cpu_per_job)

        if isinstance(self.mem_per_job, str):
            self.mem_per_job = int(self.mem_per_job)

    def __repr__(self):
        return repr(self.__dict__)

    @classmethod
    def from_dict(cls, kv):
        obj = cls()
        for k in kv:
            setattr(obj, k, kv[k])

        return obj


def get_object():
    """
    Get a Config object representing current Fiber config

    :returns: a Config object
    """
    # When an config object is needed, call this method to get an
    # concrete Fiber config object

    global _current_config

    return Config.from_dict(get_dict())


def get_dict():
    """
    Get current Fiber config in a dictionary

    :returns: a Python dictionary with all the current Fiber configurations
    """
    global_vars = globals()

    return {k: global_vars[k] for k in vars(_current_config)}


def init(**kwargs):
    """
    Init Fiber system and set config values.

    :param kwargs: If kwargs is not None, init Fiber system with corresponding
        key/value pairs in kwargs as config keys and values.

    :returns: A list of config keys that was updated in this function call.
    """
    _config = Config()

    # handle overwrites
    _config.__dict__.update(kwargs)

    updates = []
    global_vars = globals()

    for k in vars(_config):
        # fine diffs and write them
        val = getattr(_config, k)
        if k not in global_vars or global_vars[k] != val:
            global_vars[k] = val
            updates.append(k)

    global_vars["_current_config"] = _config

    logger.debug("Inited fiber with config: %s", vars(_config))

    return updates



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import os
import fiber.config as fiber_config
import logging
from fiber import context
from fiber.init import init_fiber
from fiber.meta import meta


__version__ = "0.0.9"

logger = logging.getLogger('Trax')


if sys.platform == 'win32':
    raise NotImplementedError("win32 platform is not supported now")


init_fiber()

if os.environ.get("FIBER_WORKER", None) is None:
    # Only initialize logger when fiber is imported in master process. Worker
    # process will get their logger initialized later.
    # A file based logger could be created, no logging lines after
    # this in this module.
    logger.setLevel(fiber_config.log_level)


if hasattr(sys, 'ps1'):
    _in_interactive_console = True
else:
    _in_interactive_console = False


def reset():
    init_fiber()


def init(**kwargs):
    """
    Initialize Fiber. This function is called when you want to re-initialize
    Fiber with new config values and also re-init loggers.

    :param kwargs: If kwargs is not None, init Fiber system with corresponding
        key/value pairs in kwargs as config keys and values.
    """
    init_fiber(**kwargs)


_names = [x for x in dir(context._default_context) if x[0] != "_"]
globals().update((name, getattr(context._default_context, name))
                 for name in _names)
__all__ = _names + []



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
fiber.cli

This module contains functions for the `fiber` command line tool. `fiber`
command line tool can be use do to mange the workflow of running jobs on a
computer cluster.

Check [here](getting-started.md#running-with-fiber-command) for an example of
how to use this command.

"""

import os
import json
import subprocess as sp
from pathlib import Path

import click

import fiber
import fiber.core as core
from fiber.core import ProcessStatus


CONFIG = {}


def get_backend(platform):
    from fiber.kubernetes_backend import Backend as K8sBackend
    backend = K8sBackend(incluster=False)
    return backend


def find_docker_files():
    """Find all possible docker files on current directory."""
    p = Path(".")
    q = p / "Dockerfile"

    files = list(p.glob("*.docker"))

    if q.exists():
        files.append(q)

    return files


def select_docker_file(files):
    """Ask user which docker file to use and return a PurePath object."""
    num = 0
    n = len(files)

    if n > 1:
        print("Available docker files:")
        for i, f in enumerate(files):
            print(i + 1, f.name)

        while True:
            end = len(files)
            input_str = input("which docker file to use? [1-{}] ".format(end))

            try:
                num = int(input_str) - 1
                if num < 0 or num >= end:
                    raise ValueError

                break
            except (TypeError, ValueError):
                print(
                    "Invalid input: {}. Please choose from [1-{}]".format(
                        input_str, end
                    )
                )
                continue

    return files[num]


def get_default_project_gcp():
    """Get default GCP project name."""
    name = sp.check_output(
        "gcloud config list --format 'value(core.project)' 2>/dev/null",
        shell=True,
    )
    return name.decode("utf-8").strip()


def parse_file_path(path):
    parts = path.split(":")
    if len(parts) == 1:
        return (None, path)

    if len(parts) > 2:
        raise ValueError("Bad path: {}".format(path))

    return (parts[0], parts[1])


@click.command()
@click.argument("src")
@click.argument("dst")
def cp(src, dst):
    """Copy file from a persistent storage"""
    platform = CONFIG["platform"]

    parts_src = parse_file_path(src)
    parts_dst = parse_file_path(dst)

    if parts_src[0] and parts_dst[0]:
        raise ValueError(
            "Can't copy from persistent storage to persistent storage"
        )

    if parts_src[0]:
        volume = parts_src[0]
    elif parts_dst[0]:
        volume = parts_dst[0]
    else:
        raise ValueError("Must copy/to from a persistent volume")

    k8s_backend = get_backend(platform)

    job_spec = core.JobSpec(
        image="alpine:3.10",
        name="fiber-cp",
        command=["sleep", "60"],
        volumes={volume: {"mode": "rw", "bind": "/persistent"}},
    )
    job = k8s_backend.create_job(job_spec)
    pod_name = job.data.metadata.name

    print("launched pod: {}".format(pod_name))
    exitcode = os.system(
        "kubectl wait --for=condition=Ready pod/{}".format(pod_name)
    )

    """
    status = k8s_backend.get_job_status(job)
    while status == ProcessStatus.INITIAL:
        print("Waiting for pod {} to be up".format(pod_name))
        time.sleep(1)

    if status != ProcessStatus.STARTED:
        raise RuntimeError("Tempory pod failed: {}".format(pod_name))
    """

    if parts_src[0]:
        new_src = "{}:{}".format(pod_name, parts_src[1])
        new_dst = dst
    elif parts_dst[0]:
        new_src = src
        new_dst = "{}:{}".format(pod_name, parts_dst[1])

    cmd = "kubectl cp {} {}".format(new_src, new_dst)
    os.system(cmd)

    # k8s_backend.terminate_job(job)


def detect_platforms():
    commands = ["gcloud", "aws"]
    platforms = ["gcp", "aws"]
    found_platforms = []

    for i, cmd in enumerate(commands):
        try:
            sp.check_call(["which", cmd], stdout=sp.DEVNULL)
        except sp.CalledProcessError:
            continue

        found_platforms.append(platforms[i])

    return found_platforms


def prompt_choices(choices, prompt):
    num = 0
    n = len(choices)

    if n > 1:
        for i, choice in enumerate(choices):
            print(i + 1, choice)

        while True:
            end = len(choices)
            input_str = input("{}? [1-{}] ".format(prompt, end))

            try:
                num = int(input_str) - 1
                if num < 0 or num >= end:
                    raise ValueError
                break

            except (TypeError, ValueError):
                print(
                    "Invalid input: {}. Please choose from [1-{}]".format(
                        input_str, end
                    )
                )
                continue

    return choices[num]


class DockerImageBuilder:
    def __init__(self, registry=""):
        self.registry = registry

    def get_docker_registry_image_name(image_base_name):
        return image_base_name

    def build(self):
        files = find_docker_files()
        n = len(files)
        if n == 0:
            raise RuntimeError("No docker files found in current directory")

        dockerfile = select_docker_file(files)

        cwd = os.path.basename(os.getcwd())

        # Use current dir as image base name
        image_base_name = cwd
        self.image_base_name = image_base_name

        sp.check_call(
            "docker build -f {} . -t {}".format(dockerfile, image_base_name),
            shell=True,
        )

        self.image_name = "{}:latest".format(image_base_name)

        self.tag()
        self.push()

        return self.full_image_name

    def tag(self):
        self.full_image_name = self.image_name

    def push(self):
        sp.check_call(
            "docker push {}".format(self.full_image_name), shell=True,
        )

    def docker_tag(self, in_name, out_name):
        sp.check_call("docker tag {} {}".format(in_name, out_name), shell=True)


class AWSImageBuilder(DockerImageBuilder):
    def __init__(self, registry):
        self.registry = registry
        parts = registry.split(".")
        self.region = parts[-3]

    def tag(self):
        image_name = self.image_name
        full_image_name = "{}/{}".format(self.registry, self.image_name)

        self.docker_tag(image_name, full_image_name)

        self.full_image_name = full_image_name
        return full_image_name

    def need_new_repo(self):
        output = sp.check_output(
            "aws ecr describe-repositories --region {}".format(self.region),
            shell=True,
        )
        res = json.loads(output)

        if "repositories" not in res:
            return True

        repos = res["repositories"]
        for repo in repos:
            # use self.image_base_name as repo name
            if repo["repositoryName"] == self.image_base_name:
                return False

        return True

    def create_repo_if_needed(self):
        if self.need_new_repo():
            sp.check_call(
                "aws ecr create-repository --region {} --repository-name {}".format(
                    self.region, self.image_base_name
                ),
                shell=True,
            )

        return

    def push(self):
        self.create_repo_if_needed()

        try:
            super().push()
        except sp.CalledProcessError:
            raise RuntimeError(
                "Failed to push images {}. "
                "Most likely your authorization token has expired. \nPlease run "
                '"aws ecr get-login-password --region {} | docker login --username AWS --password-stdin {}" to authenticate'.format(
                    self.full_image_name, self.region, self.registry
                )
            )


class GCPImageBuilder(DockerImageBuilder):
    def __init__(self, registry="gcr.io"):
        self.registry = registry

    def tag(self):
        image_name = self.image_name
        proj = get_default_project_gcp()

        full_image_name = "{}/{}/{}".format(self.registry, proj, image_name)
        self.docker_tag(image_name, full_image_name)

        self.full_image_name = full_image_name

        return full_image_name


@click.command(context_settings=dict(ignore_unknown_options=True,))
@click.option("-a", "--attach", is_flag=True)
@click.option("--image")
@click.option("--gpu")
@click.option("--cpu")
@click.option("--memory")
@click.option("-v", "--volume")
@click.argument("args", nargs=-1)
def run(attach, image, gpu, cpu, memory, volume, args):
    """Run a command on a kubernetes cluster with fiber."""
    platform = CONFIG["platform"]
    print(
        'Running "{}" on Kubernetes cluster ({}) '.format(
            " ".join(args), platform
        )
    )

    if image:
        full_image_name = image
    else:
        if platform == "gcp":
            builder = GCPImageBuilder()
        elif platform == "aws":
            registry = CONFIG["docker_registry"]
            if not registry:
                registry = input(
                    "What docker registry do you plan to use? "
                    "AWS registry format: "
                    '"[aws_account_id].dkr.ecr.[region].amazonaws.com"\n> '
                )
            builder = AWSImageBuilder(registry)
        else:
            raise ValueError('Unknow platform "{}"'.format(platform))

        full_image_name = builder.build()

    # run this to refresh access tokens
    exitcode = os.system("kubectl get po > /dev/null")

    job_name = os.path.basename(os.getcwd())

    k8s_backend = get_backend(platform)

    job_spec = core.JobSpec(image=full_image_name, name=job_name, command=args,)
    if gpu:
        job_spec.gpu = gpu

    if cpu:
        job_spec.cpu = cpu

    if memory:
        job_spec.mem = memory

    if volume:
        volumes = {volume: {"mode": "rw", "bind": "/persistent"}}
        job_spec.volumes = volumes

    job = k8s_backend.create_job(job_spec)
    pod_name = job.data.metadata.name
    exitcode = 0

    print("Created pod: {}".format(pod_name))

    if attach:
        # wait until job is running
        """
        os.system(
            "kubectl wait --for=condition=Ready pod/{}".format(pod_name)
        )
        """

        exitcode = os.system("kubectl logs -f {}".format(pod_name))

    if exitcode != 0:
        return exitcode

    return 0


def auto_select_platform():
    platforms = detect_platforms()
    if len(platforms) > 1:
        choice = prompt_choices(
            platforms,
            "Found many providers, which provider do you want to use",
        )
    elif len(platforms) == 1:
        choice = platforms[0]
    else:
        choice = prompt_choices(
            ["gcp", "aws"], "Which provider do you want to use"
        )

    return choice


@click.group()
@click.option("-d", "--docker-registry")
@click.option("--aws", is_flag=True, help="Run commands on Amazon AWS")
@click.option(
    "--gcp", is_flag=True, help="Run commands on Google Cloud Platform"
)
@click.version_option(version=fiber.__version__, prog_name="fiber")
def main(docker_registry, aws, gcp):
    """fiber command line tool that helps to manage workflow of distributed
    fiber applications.
    """
    if docker_registry is not None:
        CONFIG["docker_registry"] = docker_registry
    else:
        CONFIG["docker_registry"] = None

    platforms = [aws, gcp]
    platform_names = ["aws", "gcp"]
    n = sum(platforms)
    if n > 1:
        raise ValueError(
            'Only one of "{}" can be set'.format(
                ", ".join(["--" + p for p in platform_names])
            )
        )

    if aws:
        CONFIG["platform"] = "aws"
    elif gcp:
        CONFIG["platform"] = "gcp"
    else:
        CONFIG["platform"] = auto_select_platform()


main.add_command(run)
main.add_command(cp)


if __name__ == "__main__":
    main()



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
*Queues and pipes* in Fiber behave the same as in multiprocessing. The
difference is that queues and pipes are now shared by multiple processes
running on different machines.  Two processes can read from and write to the
same pipe. Furthermore, queues can be shared between many processes on
different machines and each process can send to or receive from the same queue
at the same time. Fiber's queue is implemented with
[Nanomsg](https://nanomsg.org/), a high-performance asynchronous message queue
system.

We implemented two version of `SimpleQueue`: `SimpleQueuePush` and
`SimpleQueuePull`.  To decide which queue to use, you can set
`fiber.config.use_push_queue` to `True` to use `SimpleQueuePush` or `False` to
use `SimpleQueuePull`.

The difference between `SimpleQueuePush` and
`SimpleQueuePull` is whether data is push from master process to sub-processes
or data is requested by sub-processes. When `SimpleQueuePull` is used,
sub-processes will send an message to the master process to notify master
process that it is ready. And the master process will send the data to
sub-processes. In this way, data in the queue are load-balanced between
different sub-processes due to the workload of them. When `SimpleQueuePush` is
used, all the data are push to the sub-processes as soon as possible and the
data are load-balanced in a round-robin manner. This way, there is less
overhead for sending the data to the client, but there may be uneven load
between each of the sub-processes.

Example:
```python
q = fiber.SimpleQueue()
q.put(42)

# By default, `SimpleQueue` is `SimpleQueuePull`, now we switch to
# `SimpleQueuePush`
fiber.config.use_push_queue = True
q = fiber.SimpleQueue()

```

"""


import logging
import multiprocessing.connection
import multiprocessing.reduction as reduction
from multiprocessing.util import register_after_fork as mp_register_after_fork
import atexit

from fiber.process import current_process
from fiber.backend import get_backend
from fiber.socket import Socket, ProcessDevice

# define the port range that fiber can use to listen for incoming connections
MIN_PORT = 40000
MAX_PORT = 65535
logger = logging.getLogger('Trax')


_io_threads = []
_poller_threads = {}


def _clean_up():
    for t in _io_threads:
        logger.debug("cleanup thread %s", t)
        t.stop()
    for t in _io_threads:
        logger.debug("join thread %s", t)
        t.join()


class ZConnection(multiprocessing.connection._ConnectionBase):
    """A Connection class implemented with Fiber socket.

    It takes a (sock_type, dest_addr) tuple as input, creates a new connection
    which can send and recv Python objects. This ZConnection class can be
    serialized by pickle and then send to another remote process. Fiber socket
    will be reconnected after unpickle.

    :param handle: a (sock_type, dest_addr) tuple.
        `sock_type` should be a Fiber socket type. `dest_addr` should be in
        format like "tcp://127.0.0.1:9000".

    > note: ZConnection's fileno method returns a Fiber socket.
    """

    def __init__(self, handle, readable=True, writable=True):
        self._name = None
        if handle is None:
            raise ValueError("invalid socket")

        assert isinstance(handle, tuple)

        sock_type, dest_addr = handle
        self.sock_type = sock_type
        self.dest_addr = dest_addr
        self._create_handle()

        if not readable and not writable:
            raise ValueError(
                "at least one of `readable` and `writable` must be True")
        self._readable = readable
        self._writable = writable

        mp_register_after_fork(self, ZConnection._create_handle)
        atexit.register(self._close)

    def __getstate__(self):
        return {"sock_type": self.sock_type,
                "dest_addr": self.dest_addr,
                "_readable": self._readable,
                "_writable": self._writable,
                "_name": self._name}

    def __setstate__(self, state):
        self.sock_type = state['sock_type']
        self.dest_addr = state['dest_addr']
        self._readable = state['_readable']
        self._writable = state['_writable']
        self._name = state['_name']
        self._create_handle()
        mp_register_after_fork(self, ZConnection._create_handle)
        atexit.register(self._close)

    def __del__(self):
        # __del__ is not reliable, we use atexit to clean up things
        return

    def __repr__(self):
        name = (
            self._name
            if getattr(self, "_name", None) is not None
            else self.dest_addr
        )
        return '<ZConnection [{}, {}]>'.format(name, getattr(self, "_handle", None))

    def _create_handle(self):
        logger.debug("%s _create_handle called", self)

        #self._handle = context.socket(self.sock_type)
        self._handle = Socket(mode=self.sock_type)
        self._handle.connect(self.dest_addr)
        logger.debug("connect to %s", self.dest_addr)

    def _close(self):
        if self._handle:
            self._handle.close()
            self._handle = None

    def _send_bytes(self, buf):
        self._handle.send(buf)

    # TODO(jiale) _ConnectionBase's send use _ForkingPickler. Switch to
    # send_pyobj instead?

    def _recv_bytes(self, maxsize=None):
        # TODO(jiale) support maxsize

        return self._handle.recv()

    def recv(self):
        """Receive a (picklable) object"""
        #logger.debug("recv called")
        self._check_closed()
        self._check_readable()
        buf = self._recv_bytes()
        return reduction.ForkingPickler.loads(buf)

    def _poll(self, timeout):
        return bool(self._handle.poll(timeout=timeout))

    def set_name(self, name):
        self._name = name


class LazyZConnection(ZConnection):
    def __init__(self, handle, readable=True, writable=True, name=None):
        self._name = None

        if handle is None:
            raise ValueError("invalid socket")

        if isinstance(handle, tuple):
            sock_type, dest_addr = handle
            self.sock_type = sock_type
            self.dest_addr = dest_addr
            self._inited = False
            # Do not create handle
            # self._create_handle()
        else:
            raise ValueError("can't pass a socket to LazyZConnection")

        if not readable and not writable:
            raise ValueError(
                "at least one of `readable` and `writable` must be True")
        self._readable = readable
        self._writable = writable

    def _check_closed(self):
        if self._inited is False:
            self._create_handle()
            self._inited = True
        if self._handle is None:
            raise OSError("handle is closed")

    def _close(self):
        if not self._inited:
            return
        self._handle.close()

    def _send_bytes(self, buf):
        #self._handle.send_multipart([b"", buf])
        self._handle.send(buf)

    def _recv_bytes(self, maxsize=None):
        # TODO(jiale) support maxsize
        #msg = self._handle.recv_multipart()
        # msg -> b'' b'message data' (because of ROUTER)
        #data = msg[1]
        data = self._handle.recv()

        return data

    def __getstate__(self):
        return {"sock_type": self.sock_type,
                "dest_addr": self.dest_addr,
                "_readable": self._readable,
                "_writable": self._writable}

    def __setstate__(self, state):
        self.sock_type = state['sock_type']
        self.dest_addr = state['dest_addr']
        self._readable = state['_readable']
        self._writable = state['_writable']
        self._inited = False


class LazyZConnectionPipe(LazyZConnection):
    def _send_bytes(self, buf):
        self._handle.send(buf)

    def _recv_bytes(self, maxsize=None):
        # TODO(jiale) support maxsize

        return self._handle.recv()


def Pipe(duplex=True):
    """Return a pair of connected ZConnection objects.

    :param duplex: if duplex, then both read and write are allowed on each
        of the returned connection object. Otherwise, the first returned
        connection will be read-only and the second connection will be
        write-only. By default, duplex is enabled.
    """

    if duplex:
        d = ProcessDevice("rw", "rw")
    else:
        d = ProcessDevice("r", "w")
    d.start()

    if duplex:
        return (LazyZConnection(("rw", d.out_addr,)),
                LazyZConnection(("rw", d.in_addr,)))
    return (LazyZConnection(("r", d.out_addr,)),
            LazyZConnection(("w", d.in_addr,)))


class SimpleQueuePush():
    """A queue build on top of Fiber socket. It uses "w" - ("r" - "w") - "r"
    socket combination. Messages are pushed from one end of the queue to
    the other end without explicitly pulling.
    """
    def __repr__(self):
        return "SimpleQueuePush<reader:{}, writer: {}>".format(
            self._reader_addr, self._writer_addr)

    def __init__(self):
        self.done = False
        backend = get_backend()
        ip, _, _ = backend.get_listen_addr()
        # Listens on '0.0.0.0' so that it accepts connection regardless of
        # net interfaces. When connect connects, it should use the address
        # obtained from backend.get_listen_addr().
        d = ProcessDevice("r", "w")
        d.start()

        self._reader_addr = d.out_addr
        self._writer_addr = d.in_addr

        # client side
        # set reader to None because if reader is connected, Fiber socket will
        # fairly queue  messages to all readers even if this reader is
        # not reading.
        #self.reader = None
        self.reader = LazyZConnection(("r", self._reader_addr,))
        self.writer = LazyZConnection(("w", self._writer_addr,))

    def get(self):
        """Get an element from this Queue.

        :returns: An element from this queue. If there is no element in the
            queue, this method will block.
        """
        """
        if self.reader is None:
            logger.debug(
                "reader is None, create ZConnection((\"r\", %s))",
                self._reader_addr
            )

            self.reader = LazyZConnection(("r", self._reader_addr,))
        """
        #data = self.reader._handle.recv()
        msg = self.reader.recv()
        #msg = reduction.ForkingPickler.loads(data)
        logger.debug("%s got %s", self, msg)
        return msg

    def put(self, obj):
        """Put an element into the Queue.

        :param obj: Any picklable Python object.
        """
        logger.debug("%s put %s", self, obj)
        #data = reduction.ForkingPickler.dumps(obj)
        self.writer.send(obj)

    '''
    def __getstate__(self):
        d = self.__dict__.copy()
        # set reader to None so that de-serialized Queue doesn't connect to
        # reader automatically. Fiber socket will fairly queue all the messages
        # to all the readers. We want to prevent this behavior.
        d["reader"] = None
        return d
    '''


#Pipe = ClassicPipe
SimpleQueue = SimpleQueuePush



import pytest
import time
import fiber
from fiber import Process
from fiber.queues import Pipe, SimpleQueue
import docker


def get_current_pid(q):
    import fiber
    pid = q.get()
    assert fiber.current_process().pid == pid


@pytest.fixture(scope="module")
def client():
    return docker.from_env()


class TestMisc():

    @pytest.fixture(autouse=True)
    def check_leak(self, request, client):
        assert fiber.active_children() == []
        if fiber.config.default_backend != "docker":
            yield
            return
        pre_workers = client.containers.list()
        yield
        post_workers = client.containers.list()
        assert pre_workers == post_workers

    def test_active_children(self):
        # Reset fiber module so other processes won't inteference with
        # current run
        import importlib
        importlib.reload(fiber)
        fiber.process._children = set()
        try:
            p1 = Process(target=time.sleep, args=(1,), name="test_active_children1")
            p1.start()
            p2 = Process(target=time.sleep, args=(1,), name="test_active_children2")
            p2.start()
            assert len(fiber.active_children()) == 2
        finally:
            p1.join()
            p2.join()

    def test_cpu_count(self):
        c = fiber.cpu_count()
        assert c > 0 and c != 0

    def test_current_process(self):
        c = fiber.current_process()
        assert c.pid > 0

    def test_current_process_nested(self):
        q = SimpleQueue()
        p = Process(target=get_current_pid, args=(q,), name="test_active_children")
        p.start()
        q.put(p.pid)
        p.join()
        assert p.exitcode == 0



import os
import time
from fiber import Pool, SimpleQueue
import fiber
import pytest
import docker
import threading
import random

import fiber.config as fiber_config


@pytest.fixture(scope="module")
def client():
    return docker.from_env()


def f(x):
    return x * x

def f2(x, y):
    return x * y

def fy(x, y=1):
    return x * x * y


def get_proc_name(x):
    import multiprocessing as mp
    proc = mp.current_process()
    return proc.name


def double_queue_worker(args):
    i, (instructions, returns) = args
    print('[worker]LOOPING', i)
    # instructions, returns = SUBPROC_QUEUES[i]
    assert instructions is not None
    assert returns is not None
    returns.put('READY {}'.format(i))
    print('[worker]PUT READY', i)

    while True:
        # quit after got one instruction. This makes sure each worker gets at
        # least one instruction.
        ins = instructions.get()
        print('[worker]GOT', ins, i)
        if ins == "QUIT":
            #print('[worker]got QUIT exiting', i)
            break

        returns.put("ACK {}".format(i))
        print('[worker]PUT ACK', i)


def sleep_worker(duration):
    time.sleep(duration)


def random_error_worker(i):
    random.seed(time.time())
    time.sleep(0.01)

    p = random.random()
    if p < 0.05:
        raise ValueError

    return i

def is_inside(p):
    x, y = random.random(), random.random()
    return x * x + y * y < 1

class TestPool():
    @pytest.fixture(autouse=True)
    def check_leak(self, request, client):
        assert fiber.active_children() == []
        if fiber.config.default_backend != "docker":
            yield
            return
        pre_workers = client.containers.list()
        yield
        post_workers = client.containers.list()
        assert pre_workers == post_workers

    def test_pool_basic(self):
        pool = Pool(2)
        res = pool.map(f, [1, 2, 3])
        pool.terminate()
        pool.join()
        assert res == [1, 4, 9]

    def test_pool_more(self):
        pool = Pool(4)

        # explicitly start workers instead of lazy start
        pool.start_workers()
        res = pool.map(f, [i for i in range(1000)])

        pool.wait_until_workers_up()

        pool.terminate()
        pool.join()
        assert res == [i**2 for i in range(1000)]

    def test_pool_apply(self):
        pool = Pool(4)
        res_async = pool.apply_async(f, (42,))
        r = res_async.get()
        assert r == (42 * 42)

        res = pool.apply(f, (36,))
        assert res == (36 * 36)

        res = pool.apply(fy, (36,), {"y": 2})
        assert res == (36 * 36 * 2)

        pool.terminate()
        pool.join()

    def test_pool_imap(self):
        pool = Pool(4)
        res_iter = pool.imap(f, [x for x in range(100)], 1)
        res = list(res_iter)
        assert res == [x * x for x in range(100)]

        res_iter = pool.imap_unordered(f, [x for x in range(100)], 1)
        res = list(res_iter)
        assert len(res) == 100
        res.sort()
        assert res == [x * x for x in range(100)]

        pool.terminate()
        pool.join()

    def test_pool_starmap(self):
        pool = Pool(4)
        res = pool.starmap(f, [(x,) for x in range(100)], 1)
        assert res == [x * x for x in range(100)]

        async_res = pool.starmap_async(f, [(x,) for x in range(100)], 1)
        res = async_res.get()
        assert res == [x * x for x in range(100)]

        pool.terminate()
        pool.join()

    def test_pool_starmap2(self):
        pool = Pool(4)
        res = pool.starmap(f2, [(x, x) for x in range(100)], 10)
        assert res == [x * x for x in range(100)]

        async_res = pool.starmap_async(f, [(x,) for x in range(100)], 10)
        res = async_res.get()
        assert res == [x * x for x in range(100)]

        pool.terminate()
        pool.join()

    def test_pool_multiple_workers_inside_one_job(self):
        old_val = fiber_config.cpu_per_job
        try:
            fiber_config.cpu_per_job = 2
            pool = Pool(4)
            # explicitly start workers instead of lazy start
            pool.start_workers()
            # wait for all the workers to start
            pool.wait_until_workers_up()

            res = pool.map(get_proc_name, [i for i in range(4)], chunksize=1)
            pool.terminate()
            pool.join()
            res.sort()
            # Two `ForkProcess-1` from 2 jobs, two `ForkProcess-2` from the first job
            assert res == ['ForkProcess-1', 'ForkProcess-1', 'ForkProcess-2', 'ForkProcess-2'], res
        finally:
            fiber_config.cpu_per_job = old_val

    def test_chunk_size(self):
        log_file = os.path.join(os.path.dirname(__file__), "..", "logs", "chunk.log")
        try:
            fiber.reset()
            log_file = fiber_config.log_file
            log_level = fiber_config.log_level

            fiber.init(cpu_per_job=8, log_file=log_file, log_level="debug")

            n_envs = 9
            queues = [(SimpleQueue(), SimpleQueue()) for _ in range(n_envs)]
            pool = Pool(n_envs)

            # explicitly start workers instead of lazy start
            pool.start_workers()
            print("waiting for all workers to be up")
            # wait some time for workers to start
            pool.wait_until_workers_up()
            #time.sleep(20)
            print("all workers are up")

            def run_map():
                print('[master]RUN MAP')
                # Not setting chunk size 1 here, if chunk size is calculated
                # wrong,  map will get stuck
                pool.map(double_queue_worker, enumerate(queues), chunksize=1)
                print('[master]RUN MAP DONE')

            td = threading.Thread(target=run_map, daemon=True)
            td.start()

            print('Checking...')
            for i, (_, returns) in enumerate(queues):
                print('[master]Checking queue', i, n_envs)
                assert 'READY' in returns.get()
                print(f'[master]Checking queue {i} done')

            print('All workers are ready, put HELLOs')
            for i, (instruction, _) in enumerate(queues):
                instruction.put("HELLO")
                print(f'[master]PUT HELLO {i}')

            print('All HELLOs sent, waiting for ACKs')
            for i, (_, returns) in enumerate(queues):
                assert 'ACK' in returns.get()
                print(f'[master]GOT ACK {i}')

            print('All ACKs sent, send QUIT to workers')
            for i, (instruction, _) in enumerate(queues):
                instruction.put("QUIT")
                print(f'[master]PUT QUIT {i}, {n_envs}')
            pool.terminate()
            pool.join()

        finally:
            fiber.init(log_file=log_file, log_level=log_level)

    def test_pool_close(self):
        pool = Pool(2)
        res = pool.map(f, [1, 2, 3])
        assert res == [1, 4, 9]
        pool.close()

        with pytest.raises(ValueError):
            pool.map(f, [1, 2, 3])

        pool.join()

    def test_many_jobs(self):
        """
        This is to test a race condition in handling data in pending table
        """
        workers = 5
        pool = fiber.Pool(workers)
        tasks = 5000
        duration = 0.001

        # explicitly start workers instead of lazy start
        pool.start_workers()

        pool.wait_until_workers_up()

        res = [None] * workers
        for i in range(tasks // workers):
            for j in range(workers):
                handle = pool.apply_async(sleep_worker, (duration,))
                res[j] = handle
            for j in range(workers):
                res[j].get()

        pool.terminate()
        pool.join()

    def test_pi_estimation(self):
        pool = Pool(processes=4)
        NUM_SAMPLES = int(1e6)
        pi = 4.0 * sum(pool.map(is_inside, range(0, NUM_SAMPLES))) / NUM_SAMPLES
        assert 3 < pi and pi < 4
        print("Pi is roughly {}".format(pi))

        pool.terminate()
        pool.join()

    def test_error_handling(self):
        try:
            pool = fiber.Pool(3, error_handling=True)
            # explicitly start workers instead of lazy start
            pool.start_workers()
            pool.wait_until_workers_up()
            res = pool.map(
                random_error_worker, [i for i in range(300)], chunksize=1
            )

            assert res == [i for i in range(300)]

        finally:
            pool.terminate()
            pool.join()

    def test_error_handling_unordered(self):
        try:
            pool = fiber.Pool(3, error_handling=True)
            # explicitly start workers instead of lazy start
            pool.start_workers()
            pool.wait_until_workers_up()
            res_iter = pool.imap_unordered(
                random_error_worker, [i for i in range(300)], chunksize=1
            )

            res = list(res_iter)
            res.sort()

            assert res == [i for i in range(300)]

        finally:
            pool.terminate()
            pool.join()

    def test_pool_with_no_argument(self):
        # Make sure no exception is raised
        p = fiber.Pool()
        p.map(print, [1, 2, 3, 4])
        p.terminate()
        p.join()

        assert 1 == 1



import os
import re
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys
from subprocess import check_call

install_requires = [
    "nnpy-bundle",
    "psutil",
    "docker <= 4.2.2",
    "cloudpickle",
    "kubernetes",
    "requests",
    "click",
]

tests_require = [
    "pytest",
    "docker <= 4.2.2",
    "flake8",
]

extras = {
    "test": tests_require,
}


def find_version(*file_paths):
    basedir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(basedir, *file_paths)) as fp:
        content = fp.read()
        version_match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M
        )
        if version_match:
            return version_match.group(1)

        raise RuntimeError("Version string not found.")


def read_long_description(filename="README.md"):
    with open(filename) as f:
        return f.read().strip()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


setup(
    author="Jiale Zhi",
    author_email="jiale@uber.com",
    description="A distributed computing library for modern computer clusters",
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras,
    tests_require=tests_require,
    cmdclass={"test": PyTest},
    license="Apache 2.0",
    long_description=read_long_description(),
    long_description_content_type="text/markdown",
    name="traxfiber",
    packages=find_packages(),
    entry_points={
        "console_scripts": ["fiber=fiber.cli:main"],
    },
    version=find_version("fiber", "__init__.py"),
    zip_safe=False,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Distributed Computing",
    ],
    url="https://github.com/trax-retail/trax-fiber",
)



import os
import fiber.backend
import fiber.config
import fiber.docker_backend
import fiber.kubernetes_backend
import fiber.local_backend


class TestBackend():
    def test_backend_selection(self):
        backend = fiber.backend.auto_select_backend()
        assert backend == fiber.config.default_backend

        try:
            orig_val = os.environ.get("KUBERNETES_SERVICE_HOST", "")
            os.environ["KUBERNETES_SERVICE_HOST"] = "127.0.0.1"

            backend = fiber.backend.auto_select_backend()
            assert backend == "kubernetes"
        finally:
            os.environ["KUBERNETES_SERVICE_HOST"] = orig_val

    def test_backend_initialization(self):
        doker_backend = fiber.backend.get_backend("docker")
        assert isinstance(doker_backend, fiber.docker_backend.Backend)
        assert doker_backend.name == "docker"

        # kuberntes backend needs k8s config, so it's skipped here
        #k8s_backend = fiber.backend.get_backend("kubernetes", incluster=False)
        #assert isinstance(k8s_backend, fiber.kubernetes_backend.Backend)
        #assert k8s_backend.name == "kubernetes"

        local_backend = fiber.backend.get_backend("local")
        assert isinstance(local_backend, fiber.local_backend.Backend)
        assert local_backend.name == "local"
